// *** PocketMotus Windows CE ***
// Version 1.1
// Franois PESSAUX June 2001
// This program is freeware.
// Use its sources as you whant provided that you leave this copyright unchanged
// and that you drop a small word telling that I'm the author of these sources.
// Thank's a lot.
//   --Francois


#include "stdafx.h"
#include "resource.h"


// Automatically call RefreshScreen as a side effect.
void ResizeBoardWindow (HWND w_handle, LPARAM lparam_from_wmsettingchanged)
{
	SIPINFO si ;
	TEXTMETRIC text_metrix ;
	HDC disp_dev_handle ;
	SCROLLINFO scroll_info ;
	RECT rect ;
	int scrollbar_compensation ;
	int iDelta, new_height, new_width ;

	// Get the SIP status info to know the visible area on the screen.
	si.cbSize = sizeof (si) ;
	// In the doc, there is written about iuParam (2 argument of SHSipInfo):
	//		"To improve the response time of SHSipInfo, set uiParam to the lParam
	//			of WM_SETTINGCHANGE."
	// Indeed, the correct formulation should TO MAKE IT WORKING, set ...
	// If uiParam is not assigned to l_param, the call with SPI_GETSIPINFO fails
	// to report the correct rcVisibleDesktop field of the SIPINFO structure.
	// So the mess everything !
	SHSipInfo (SPI_GETSIPINFO, lparam_from_wmsettingchanged, &si, 0) ;
	iDelta = (si.fdwFlags & SIPF_ON) ? 0 : MENU_HEIGHT ;

	// Get the current font size to compute the size of cases containing letters.
	disp_dev_handle = GetWindowDC (w_handle) ;
	GetTextMetrics (disp_dev_handle, &text_metrix) ;
	ReleaseDC (w_handle, disp_dev_handle) ;

	// Get the current font size to compute the size of cases containing letters.
	case_width = text_metrix.tmMaxCharWidth ;
	case_height = text_metrix.tmHeight ;

	// Note that SM_CXFRAME is not supported by Windows CE, so we add manuall 2
	// to compensate the borders of the scrollbar.
	scrollbar_compensation = 2 + GetSystemMetrics (SM_CXVSCROLL) ;

	// Now resize the board window.
	GetWindowRect (main_dialog_hd, &rect) ;		// Rect contents is relative to the top/left of the screen.
	new_width = scrollbar_compensation +
		(case_width * WORD_LEN) +
		(WORD_LEN + 1) * CHECKER_BORDER +
		40 ;		// 40: Space left to display the turn number.
	new_height = si.rcVisibleDesktop.bottom - rect.bottom - iDelta - TOP_BORDER ;

	// Compute the number of turns fitting in the window or the window height
	// if there is less turns to display than there is room on the screen.
	number_of_turns_to_display =
		(new_height - CHECKER_BORDER) / (case_height + CHECKER_BORDER) ;
	number_of_turns_to_display = min (number_of_turns_to_display, number_of_turns) ;

	// Check and correct he first visible line in case where changing the window's
	// size caused the first visible line to change (indeed to be upper that it
	// was before).
	if (current_visible_start_turn < 0)
		current_visible_start_turn = 0 ;
	if (current_visible_start_turn >= number_of_turns - number_of_turns_to_display)
		current_visible_start_turn = number_of_turns - number_of_turns_to_display ;

	// Adjust the height to fit EXACTLY the number of turns to display.
	new_height = case_height * number_of_turns_to_display +
		(number_of_turns_to_display + 1) * CHECKER_BORDER ;

	// Be careful, we move the window relatively to the parent window.
	// And the main_dialog overlaps this parent window, so let's correct coordinates.
	// By the way, ask to perform a WM_PAINT immediately, so that it will automatically
	// call RefreshBoardWindow, and finish tuning the REAL size of the window and
	// the real number of turns to be actually displayed. For this reason, the last
	// parameter of MoveWindow is set to TRUE !
	MoveWindow (w_handle, LEFT_BORDER, (rect.bottom - MENU_HEIGHT) + TOP_BORDER, new_width, new_height, TRUE) ;

	// Compute the scrollbar attributes.
	page_vertical_scroll_size =	max (0, (number_of_turns_to_display)) ;

	// Set the scrollbar attributes.
	scroll_info.cbSize = sizeof (scroll_info) ;
  scroll_info.nMin = 0 ;
	// scroll_info.nMax = max (0, number_of_turns - number_of_turns_to_display) ;
	scroll_info.nMax = max (0, number_of_turns - 1) ;
	scroll_info.nPage = page_vertical_scroll_size ;
  scroll_info.nPos = current_visible_start_turn ;
  scroll_info.fMask = SIF_DISABLENOSCROLL|SIF_ALL ;
  SetScrollInfo (w_handle, SB_VERT, &scroll_info, TRUE) ;
}




// Used by BoardWindowHandler for screen refresh.
void RefreshBoardWindow (HWND w_handle)
{
	int i, j, real_i ;
	HDC disp_dev_handle ;
  PAINTSTRUCT paint_struct ;
	SIZE text_size ;
	RECT r ;
	int center_amount ;
	POINT points[2] ;
	HBRUSH old_brush ;

  BeginPaint (w_handle, &paint_struct) ;
	disp_dev_handle = GetWindowDC (w_handle) ;

	// Get the board window's size.
	GetClientRect (w_handle, &r) ;

	// Redraw the checker.
  // Draw the window's background. Don't forget to backup the old brush object
	// to restore it later to avoid memory leak !
	old_brush = (HBRUSH) SelectObject (disp_dev_handle, (HBRUSH) GetStockObject (GRAY_BRUSH)) ;
	Rectangle (disp_dev_handle, r.left, r.top, r.right - r.left, r.bottom - r.top) ;
	// Restore the previous brush object to avoid memory leak .
	SelectObject (disp_dev_handle, old_brush) ;

	for (i=0; i<=number_of_turns_to_display; i++) {
		// Draw the horizontal lines of the checker.
		points[0].x = 0 ;
		points[0].y = (case_height + CHECKER_BORDER) * i ;
		points[1].x = (case_width + CHECKER_BORDER) * WORD_LEN + CHECKER_BORDER ;
		points[1].y = points[0].y ;
		Polyline (disp_dev_handle, points, 2) ;
		}

	for (i=0; i<=WORD_LEN; i++) {
		// Draw the vertical lines of the checker.
		points[0].x = (case_width + CHECKER_BORDER) * i ;
		points[0].y = 0 ;
		points[1].x = points[0].x ;
		points[1].y = (case_height + CHECKER_BORDER) * number_of_turns + CHECKER_BORDER ;
		Polyline (disp_dev_handle, points, 2) ;
		}

	// Redraw the words if a game exists.
	if (current_game != NULL) {
		for (i=0; i<number_of_turns_to_display; i++) {
			// Compute the index of the displayed turn according to the scrolling.
			real_i = i + current_visible_start_turn ;
			for (j=0; j<WORD_LEN; j++) {
				// Just debug ;-)
				if (real_i >= number_of_turns) {
					MessageBox (w_handle,
						TEXT ("Display misfeature. Please report !"),
						TEXT ("Debug warning"), 
						MB_OK|MB_ICONWARNING|MB_APPLMODAL|MB_SETFOREGROUND) ;
					}
				SetBkColor (disp_dev_handle, current_game->colors_history[real_i][j]) ;
				swprintf (tmp_unicode_string, TEXT ("%C"), current_game->suggestions_history[real_i][j]) ;
				GetTextExtentPoint (disp_dev_handle, tmp_unicode_string, 1, &text_size) ;
				center_amount = (case_width - text_size.cx) / 2 ;

				r.left = (case_width + CHECKER_BORDER) * j + CHECKER_BORDER ;
				r.top = (case_height + CHECKER_BORDER) * i + CHECKER_BORDER ;
				r.right = r.left + case_width ;
				r.bottom = r.top + case_height ;
				ExtTextOut (disp_dev_handle, r.left + center_amount, r.top, ETO_OPAQUE, &r,
					tmp_unicode_string, 1, NULL) ;
				}																			// End of for (j=0; j<WORD_LEN; j++)

			// Now display the turn number (starting from 1, instead of 0).
			j = swprintf (tmp_unicode_string, TEXT ("# %d"), real_i +1) ;		// Reuse j to save a variable.
			GetTextExtentPoint (disp_dev_handle, tmp_unicode_string, j, &text_size) ;
			r.left = r.left + case_width + 10 ;			// 10: Space left between the lat letter and the turn number.
			r.right = r.left + text_size.cx ;
			// Let's use a simple white background.
			SetBkColor (disp_dev_handle, RGB (255, 255, 255)) ;
			ExtTextOut (disp_dev_handle, r.left, r.top, ETO_OPAQUE, &r,
				tmp_unicode_string, j, NULL) ;
			}																				// End of for (i=0; i<number_of_turns_to_display; i++)
		}																					// End of if (current_game != NULL)
	
	ReleaseDC (w_handle, disp_dev_handle) ;
  EndPaint (w_handle, &paint_struct) ;
}



void UpdateGame (void) ;
void CALLBACK GlobalTimerProc (HWND hwnd, UINT uMsg, UINT idEvent, DWORD dwTime)
{
	// In any case, attempt to kill the timers if some exists
	// Because this function is a callback called automagically,
	// UpdateGame is only called after, and then the timers are not yet
	// disabled here. So do it now !
	KillTimer (main_window_handle, STEP_TIMER_ID) ;
	KillTimer (main_window_handle, GLOBAL_TIMER_ID) ;
	// Set the current suggestion to "nothing".
	strcpy (current_game->suggestion, "******") ;
	MessageBox (main_window_handle,
			TEXT ("Not fast enough. You lost this turn..."),
			TEXT ("Information"), 
			MB_OK|MB_ICONINFORMATION|MB_APPLMODAL|MB_SETFOREGROUND) ;
	UpdateGame () ;
}



void CALLBACK StepTimerProc (HWND hwnd, UINT uMsg, UINT idEvent, DWORD dwTime)
{
	// Set the current suggestion to "nothing".
	remaining_time-- ;
	swprintf (tmp_unicode_string, TEXT ("%d s"), remaining_time) ;
	SetWindowText (main_timer_hd, tmp_unicode_string) ;
}



// Terminate a game without freeing it.
void TerminateGame ()
{
	HMENU menu_handle ;

	if (current_game != NULL) {
		current_game->current_turn = -1 ;
		// Disable menus, button and edit controls.
		EnableWindow (main_edit_hd, FALSE) ;
		menu_handle = (HMENU) SendMessage (main_menubar_hd, SHCMBM_GETMENU, (WPARAM)0, (LPARAM) 0) ;
		EnableMenuItem (menu_handle, IDMI_GAME_ABORTGAME, MF_GRAYED) ;
		EnableMenuItem (menu_handle, IDMI_SOFT_HINT, MF_GRAYED) ;
		EnableMenuItem (menu_handle, IDMI_HARD_HINT, MF_GRAYED) ;
		EnableWindow (GetDlgItem (main_dialog_hd, IDC_BUTTON_SUBMIT), FALSE) ;
		}
}



void UpdateGame ()
{
	BOOL seen[WORD_LEN] ;
	char place[WORD_LEN] ;
	int i, j ;

	// In any case, attempt to kill the timers if some exist.
	KillTimer (main_window_handle, STEP_TIMER_ID) ;
	KillTimer (main_window_handle, GLOBAL_TIMER_ID) ;

	// Get the current suggestion.
	strcpy (current_game->suggestions_history[current_game->current_turn],
		current_game->suggestion) ;

	// Check for the end of the game.
	if (strcmp (current_game->suggestion, current_game->solution) == 0) {
		// The player just won.
		MessageBox (main_window_handle, TEXT ("Congratulation, you won !"),
			TEXT ("Game over"), 
			MB_OK|MB_ICONINFORMATION|MB_APPLMODAL|MB_SETFOREGROUND) ;
		// Just set all the letters with a green background to show that they
		// were all guessed right.
		for (i=0; i<WORD_LEN; i++) {
			current_game->colors_history[current_game->current_turn][i] = GOOD_PLACE_COLOR ;
			current_game->suggestions_history[current_game->current_turn][i] = current_game->solution[i] ;
			}
		// And then terminate the game.
		TerminateGame () ;
		return ;
		}

	// The player just lost the game.
	if (current_game->current_turn == (number_of_turns - 1)) {
		swprintf (tmp_unicode_string,
			TEXT ("So bad ! The solution was '%S'."), current_game->solution) ;
		MessageBox (main_window_handle, tmp_unicode_string, TEXT ("Game over"), 
			MB_OK|MB_ICONINFORMATION|MB_APPLMODAL|MB_SETFOREGROUND) ;
		TerminateGame () ;
		return ;
	}

	// Initiate the checking process.
	for (i=0; i<WORD_LEN; i++) {
		seen[i] = FALSE ;
		place[i] = WRONG_LETTER ;
		}

	// First check if the player want to play hard by losing his turn
	// if his suggestion is not a valid word inside the dictionary.
	if (check_flag) {
		// If the suggestion does not exist then the player loses his turn.
		if (!Search ((unsigned char*) current_game->suggestion)) {
			strcpy (current_game->suggestion, "******") ;
			strcpy (current_game->suggestions_history[current_game->current_turn],
				current_game->suggestion) ;
			MessageBox (main_window_handle,
				TEXT ("Your suggestion is not in the dictionary. You lost this turn..."),
				TEXT ("Information"), 
				MB_OK|MB_ICONINFORMATION|MB_APPLMODAL|MB_SETFOREGROUND) ;
			goto lose_turn ;
		}
	}

	// Else we check for the position of the letters.
	for (i=0; i<WORD_LEN; i++) {
		if (current_game->solution[i] == current_game->suggestion[i]) {
			seen[i] = TRUE ;
			place[i] = GOOD_PLACE ;
		}
	}

	for (i=0; i<WORD_LEN; i++) {
		if (place[i] != GOOD_PLACE) {
			for (j=0; j<WORD_LEN; j++) {
				if ((!seen[j]) && (current_game->suggestion[i] == current_game->solution[j])) {
					place[i] = BAD_PLACE ;
					seen[j] = TRUE ;
					break ;
				}
			}
		}
	}

	lose_turn:
	for (i=0; i<WORD_LEN; i++) {
		switch (place[i]) {
			case WRONG_LETTER:
				current_game->colors_history[current_game->current_turn][i] = WRONG_LETTER_COLOR ;
				swprintf (tmp_unicode_string, TEXT ("%C"), current_game->suggestion[i]) ;
				break ;												// End of WRONG_LETTER

			case BAD_PLACE:
				current_game->colors_history[current_game->current_turn][i] = BAD_PLACE_COLOR ;
				swprintf (tmp_unicode_string, TEXT ("%C"), current_game->suggestion[i]) ;
				break ;												// End of BAD_PLACE

			case GOOD_PLACE:
				// Place the letter on the next line, that is give the hint for the next turn.
				current_game->suggestions_history[current_game->current_turn+1][i] = current_game->solution[i] ;
				current_game->colors_history[current_game->current_turn+1][i] = GOOD_PLACE_COLOR ;
				// Now manage the current line's color and letter.
				current_game->colors_history[current_game->current_turn][i] = GOOD_PLACE_COLOR ;
				swprintf (tmp_unicode_string, TEXT ("%C"), current_game->solution[i]) ;
				break ;												// End of GOOD_PLACE
			}																// End of switch (place[i])
		}																	// End of for (i=0; i<WORD_LEN; i++)

	// Now merge the known letters we got from previous turns and the known letters
	// we just learnt from this turn.
	for (i=0; i<WORD_LEN; i++) {
		if (current_game->current_known_facts[i] != ' ') {
			current_game->suggestions_history[current_game->current_turn+1][i] = current_game->current_known_facts[i] ;
			current_game->colors_history[current_game->current_turn+1][i] = GOOD_PLACE_COLOR ;
			}																// End of if (current_known_facts[i] != " ")
		else {
				if (current_game->colors_history[current_game->current_turn+1][i] == GOOD_PLACE_COLOR)
					current_game->current_known_facts[i] = current_game->suggestions_history[current_game->current_turn+1][i] ;
			}																// End of else of if (current_known_facts[i] != " ")
		}																	// End of for (i=0; i<WORD_LEN; i++)

	// Next turn.
	current_game->current_turn++ ;

	// Scroll to next line if needed and set the focus to the edit control.
	if (current_game->current_turn >= number_of_turns_to_display)
		SendMessage (board_window_handle, WM_VSCROLL, MAKEWPARAM (SB_LINEDOWN, 0), (LPARAM) NULL) ;
	SetFocus (main_edit_hd) ;

	// Force screen refresh.
	InvalidateRect (board_window_handle, NULL, FALSE) ;

	// Set timer again if needed.
	if (timer_flag) {
		remaining_time = timer_limit ;
		SetTimer (main_window_handle, STEP_TIMER_ID, 1 * 1000, StepTimerProc) ;
		SetTimer (main_window_handle, GLOBAL_TIMER_ID,
			timer_limit * 1000, GlobalTimerProc) ;
		}																	//End of f (timer_flag)
	else SetWindowText (main_timer_hd, TEXT ("")) ;
	// Clear the user's suggestion.
	SetWindowText (main_edit_hd, TEXT ("")) ;
}



void ReleaseGame ()
{
	int i ;

	// First, kill all timers just in case...
	KillTimer (main_window_handle, STEP_TIMER_ID) ;
	KillTimer (main_window_handle, GLOBAL_TIMER_ID) ;

	// Now, free the used memory.
	if (current_game != NULL) {
		if (current_game->suggestions_history) {
			for (i=0; i<number_of_turns; i++) {
				if (current_game->suggestions_history[i])	free (current_game->suggestions_history[i]) ;
				}
			free (current_game->suggestions_history) ;
			}

		if (current_game->colors_history) {
			for (i=0; i<number_of_turns; i++) {
				if (current_game->colors_history[i]) free (current_game->colors_history[i]) ;
				}
			free (current_game->colors_history) ;
			}

		free (current_game) ;
		}

	current_game = NULL ;
}




void StartNewGame ()
{
	int i, j ;
	HMENU menu_handle ;

	// Ask for current game abortion only if a game is running.
	if ((current_game != NULL) && (current_game->current_turn != -1)) {
		if (MessageBox (main_window_handle,
					TEXT ("Abort current game and start a new one ?"),
					TEXT ("Want more ?"),
					MB_YESNO|MB_ICONQUESTION|MB_APPLMODAL|MB_SETFOREGROUND) != IDYES)
			return ;
			}

	ReleaseGame () ;
	current_game = (struct game*) malloc (sizeof (struct game)) ;
	current_game->current_turn = 0 ;
	if (!RandomChoose (current_game->solution)) {
		MessageBox (main_window_handle,
			TEXT ("Unable to choose a word. May be your dictionary is absent or corrupted."),
			TEXT ("Error"), MB_OK|MB_ICONERROR|MB_APPLMODAL|MB_SETFOREGROUND) ;
		ReleaseGame () ;
		return ;
		}

	current_game->suggestions_history = (char**) malloc (number_of_turns * sizeof (char*)) ;
	current_game->colors_history = (COLORREF**) malloc (number_of_turns * sizeof (COLORREF*)) ;
	for (i=0; i<number_of_turns; i++) {
		current_game->suggestions_history[i] = (char*) malloc ((1 + WORD_LEN) * sizeof (char)) ;
		current_game->colors_history[i] = (COLORREF*) malloc (WORD_LEN * sizeof (COLORREF)) ;
		for (j=0; j<WORD_LEN; j++) current_game->colors_history[i][j] = WRONG_LETTER_COLOR ;
		strcpy (current_game->suggestions_history[i], BLANK_STRING_INIT) ;
		}

	// Set the first line's color and starting/ending letters, that is the initial hint,
	// and update the currently known facts (i.e the currently known letters at the
	// right place.
	strcpy (current_game->current_known_facts, BLANK_STRING_INIT) ;
	current_game->current_known_facts[0] = current_game->solution[0] ;
	current_game->current_known_facts[WORD_LEN-1] = current_game->solution[WORD_LEN-1] ;
	strcpy (current_game->suggestions_history[0], current_game->current_known_facts) ;
	current_game->colors_history[0][0] = GOOD_PLACE_COLOR ;
	current_game->colors_history[0][WORD_LEN-1] = GOOD_PLACE_COLOR ;

	// Force screen refresh.
	InvalidateRect (board_window_handle, NULL, FALSE) ;

	// Initialize the timer if needed.
	if (timer_flag) {
		remaining_time = timer_limit ;
		SetTimer (main_window_handle, STEP_TIMER_ID, 1 * 1000, StepTimerProc) ;
		SetTimer (main_window_handle, GLOBAL_TIMER_ID,
			timer_limit * 1000, GlobalTimerProc) ;
	}
	else SetWindowText (main_timer_hd, TEXT ("")) ;
	
	// Enable the edit control and clear it and set the scrollbar at the first line.
	EnableWindow (main_edit_hd, TRUE) ;
	SetWindowText (main_edit_hd, TEXT ("")) ;
	// Set the scrolling window onto th efirst line's turn.
	SendMessage (board_window_handle, WM_VSCROLL, MAKEWPARAM (SB_TOP, 0), (LPARAM) NULL) ;
	// Enable menus, button and edit controls.
	menu_handle = (HMENU) SendMessage (main_menubar_hd, SHCMBM_GETMENU, (WPARAM)0, (LPARAM) 0) ;
	EnableMenuItem (menu_handle, IDMI_GAME_ABORTGAME, MF_ENABLED) ;
	EnableMenuItem (menu_handle, IDMI_SOFT_HINT, MF_ENABLED) ;
	EnableMenuItem (menu_handle, IDMI_HARD_HINT, MF_ENABLED) ;
	EnableWindow (GetDlgItem (main_dialog_hd, IDC_BUTTON_SUBMIT), TRUE) ;
}



// Modeless Dialog box handler.
BOOL CALLBACK MainDlgHandler (HWND w_handle, UINT message, WPARAM w_param,
															LPARAM l_param)
{
	int wm_id, wm_event ;

	wm_id    = LOWORD (w_param) ;
  wm_event = HIWORD (w_param) ;

	switch (message) {
		case WM_INITDIALOG:
			main_edit_hd = GetDlgItem (w_handle, IDC_EDIT_MAIN) ;
			// Set the limit of the entry box
			SendMessage (main_edit_hd, EM_SETLIMITTEXT, 6, 0) ;
			main_timer_hd = GetDlgItem (w_handle, IDC_STATIC_TIMER) ;
			return (TRUE) ;
			break ;															// End of WM_INITDIALOG

		case WM_COMMAND:
			switch (wm_id) {
				case IDC_EDIT_MAIN:
					switch (wm_event) {
						case EN_SETFOCUS:
							// Use the undocummented flag SIPF_DISABLECOMPLETION to avoid passords to
							// be stored in the word completion dictionary !!!
							SipShowIM (SIPF_ON|SIPF_DISABLECOMPLETION) ;
							// Force the system to finish handling the message.
							return (FALSE) ;
							break ;												// End of EN_SETFOCUS

						/* No, I think it's better to keep the input panel open...
						case EN_KILLFOCUS:
							SipShowIM (SIPF_OFF) ;
							return (FALSE) ;
							break ;											// End of EN_KILLFOCUS
						*/
						}															// End of switch (wm_event)
					break ;													// End of IDC_EDIT_MAIN

				case IDC_BUTTON_SUBMIT:
					GetWindowText (main_edit_hd, tmp_unicode_string, WORD_LEN+1) ;
					if (wcslen (tmp_unicode_string) != WORD_LEN) {
						MessageBox (w_handle,
							TEXT ("Your suggestion must be 6 characters long !"), TEXT ("Error"),
							MB_OK|MB_ICONERROR|MB_APPLMODAL|MB_SETFOREGROUND) ;
						}															// End of if (wcslen (tmp_unicode_string) != WORD_LEN)
					else {
						wcstombs ((char*) tmp_char_string, tmp_unicode_string, WORD_LEN+1) ;
						strcpy (current_game->suggestion, tmp_char_string) ;
						UpdateGame () ;
						}													// End of else of if (wcslen (tmp_unicode_string) != WORD_LEN)
					return (TRUE) ;
					break ;											// End of IDC_BUTTON_SUBMIT
				}															// End of switch (wm_id)
			break ;													// End of WM_COMMAND
		}																	// End of switch (message)

	// Force default handling by the system (because we're
	// in the case of a dialog handler). Calling DefWindowProc is only the
	// protocol for window handlers.
	return (FALSE) ;
}




LRESULT CALLBACK BoardWindowHandler (HWND w_handle, UINT message, WPARAM w_param,
																		LPARAM l_param)
{
  int wm_id, wm_event ;

  wm_id    = LOWORD (w_param) ;
  wm_event = HIWORD (w_param) ;

	switch (message) {
		case WM_CREATE:
			ResizeBoardWindow (w_handle, (LPARAM) 0) ;
			break ;															// End of WM_CREATE

		case WM_PAINT:
			// This call handles Begin/EndPaint itself, so no need to do it here.
			RefreshBoardWindow (w_handle) ;
			break ;															// End of WM_PAINT


		case WM_VSCROLL: {
			SCROLLINFO si ;
			int old_visible_start_turn ;

			old_visible_start_turn = current_visible_start_turn ;
			switch (wm_id) {
				case SB_LINEUP:
					current_visible_start_turn-- ;
					break ;													// End of SB_LINEUP

				case SB_LINEDOWN:
					current_visible_start_turn++ ;
					break ;													// End of SB_LINEDOWN

				case SB_PAGEUP:
					current_visible_start_turn -= (page_vertical_scroll_size - 1) ;
					break ;													// End of SB_PAGEUP

				case SB_PAGEDOWN:
					current_visible_start_turn += (page_vertical_scroll_size - 1) ;
					break ;													// End of SB_PAGEDOWN

				case SB_THUMBPOSITION:
					current_visible_start_turn = wm_event ;
					break ;													// End of SB_THUMBPOSITION

				case SB_TOP:
					current_visible_start_turn = 0 ;
					break ;													// End of SB_TOP
				}																	// End of switch (wm_id)

			// Check and correct range because pressing the arrows up down even if
			// the scrollbar is already blocked *does* send the SB_LINEUP/SB_LINEDOWN
			// message. Indeed there's no link between the fact that one presses the
			// arrows and the nMax value of the scrollbar. This means that nMax is the
			// maximum value the scrollbar can return, but even if the scrollbar is
			// blocked, pressing the arrow generates always a SB_LINExx message !
			if (current_visible_start_turn < 0)
				current_visible_start_turn = 0 ;
			if (current_visible_start_turn >= number_of_turns - number_of_turns_to_display)
				current_visible_start_turn = number_of_turns - number_of_turns_to_display ;

			// If scroll position changed, update scrollbar and force redraw of window.
			if (current_visible_start_turn != old_visible_start_turn) {
        si.cbSize = sizeof (si) ;
				si.nPos = current_visible_start_turn ;
        si.fMask = SIF_DISABLENOSCROLL|SIF_POS ;
        SetScrollInfo (w_handle, SB_VERT, &si, TRUE) ;
        InvalidateRect (w_handle, NULL, FALSE) ;
			}																	// End of if (current_visible_start_y != old_visible_start_y)
			break ;														// End of WM_VSCROLL
			}																	// End of local block for WM_SCROLL
		}																		// End of switch (message)
	
	// Remember, window handler must call DefWindowProc to handle
	// the remaining messages. Returning TRUE or FALSE is only the protocol.
	// for dialog handlers !
	return (DefWindowProc (w_handle, message, w_param, l_param)) ;
}




void GiveHint (int hint_type)
{
	int i ;

	if (current_game != NULL) {
		// Exclude first and last letters because they are given as the game begins.
		i = (rand () % (WORD_LEN - 2)) + 1 ;
		switch (hint_type) {
			case IDMI_SOFT_HINT:
				// Asked for a soft hint.
				swprintf (tmp_unicode_string, TEXT ("I propose you a '%C'. Now place it at the right place..."), current_game->solution[i]) ;
				break ;														// End of IDMI_SOFT_HINT

			case IDMI_HARD_HINT:
				// Asked for a hard hint. Number cases from 1 to word's length instead of 0.
				swprintf (tmp_unicode_string, TEXT ("I propose you a '%C' in case number %d."), current_game->solution[i], i + 1) ;
				break ;														// End of IDMI_HARD_HINT
			}																		// End of switch (hint_type)

		MessageBox (main_window_handle,	tmp_unicode_string, TEXT ("Information"),
			MB_OK|MB_ICONINFORMATION|MB_APPLMODAL|MB_SETFOREGROUND) ;
		}																			// End of if (current_game != NULL)
	else {
		MessageBox (main_window_handle, TEXT ("No game currently running ... so no hint needed !"), TEXT ("Error"),
			MB_SETFOREGROUND|MB_OK|MB_ICONERROR|MB_APPLMODAL) ;
		}																			// End of else of if (current_game != NULL)
}



LRESULT CALLBACK OptionDialogHandler (HWND w_handle, UINT message,
									  WPARAM w_param, LPARAM l_param)
{
	int wm_id, wm_event ;
	
	wm_id    = LOWORD (w_param) ;
	wm_event = HIWORD (w_param) ;
	
	switch (message) {
		case WM_INITDIALOG:
			// Set the current dictionary name.
			SetWindowText (GetDlgItem (w_handle, IDC_STATIC_DICNAME), dic_name) ;
			// Set the state of the check buttons.
			SendMessage (GetDlgItem (w_handle, IDC_CHECK_TIMER), BM_SETCHECK, (WPARAM) timer_flag, (LPARAM) 0) ;
			SendMessage (GetDlgItem (w_handle, IDC_CHECK_LOOKUP), BM_SETCHECK, (WPARAM) check_flag, (LPARAM) 0) ;
			// Set the time limit edit control's text.
			wsprintf (tmp_unicode_string, TEXT ("%d"), timer_limit) ;
			SetWindowText (GetDlgItem (w_handle, IDC_EDIT_TIME_LIMIT), tmp_unicode_string) ;
			// Set the number of turns for a game in the edit control.
			wsprintf (tmp_unicode_string, TEXT ("%d"), number_of_turns) ;
			SetWindowText (GetDlgItem (w_handle, IDC_EDIT_TURNS), tmp_unicode_string) ;
			return (TRUE) ;
			break ;																				// End of WM_INITDIALOG

		case WM_COMMAND:
			switch (wm_id) {
				case IDB_OPTIONSOK:
					if (MessageBox (w_handle, TEXT ("This will abort the current game and begin a new one. Continue ?"),
								TEXT ("Warning"), MB_SETFOREGROUND|MB_YESNO|MB_ICONQUESTION|MB_APPLMODAL) == IDYES) {
						// Abort the current game and disable buttons
						ReleaseGame () ;

						// Get the dictionary name
						GetWindowText (GetDlgItem (w_handle, IDC_STATIC_DICNAME), dic_name, MAX_STRINGLEN) ;
						UnloadDictionary () ;
						if (!LoadDictionary (dic_name)) {
							// Clear the current dictionary name.
							wcscpy (dic_name, TEXT ("\0")) ;
							MessageBox (w_handle, TEXT ("Unable to load dictionary. You may select a valid one later."), TEXT ("Warning"),
								MB_SETFOREGROUND|MB_OK|MB_ICONWARNING|MB_APPLMODAL) ;
							}										// End of if (!LoadDictionary (dic_name))
						// Update the dictionary name label.
						SetWindowText (GetDlgItem (w_handle, IDC_STATIC_DICNAME), dic_name) ;
						// Get the state of the 2 checkbuttons.
						timer_flag = (BOOL) (SendMessage (GetDlgItem (w_handle, IDC_CHECK_TIMER), BM_GETCHECK, (WPARAM) 0, (LPARAM) 0) == BST_CHECKED) ;
						check_flag = (BOOL) (SendMessage (GetDlgItem (w_handle, IDC_CHECK_LOOKUP), BM_GETCHECK, (WPARAM) 0, (LPARAM) 0) == BST_CHECKED) ;
						// Get the time limit edit control's text.
						GetWindowText (GetDlgItem (w_handle, IDC_EDIT_TIME_LIMIT), tmp_unicode_string, MAX_STRINGLEN) ;
						timer_limit = (int) _wtoi (tmp_unicode_string) ;
						if (timer_limit < 10) {
							timer_limit = 10 ;
							MessageBox (w_handle, TEXT ("Timer limit is lower than 10 seconds. Let me set it at 10 otherwise it will be impossible to play !"),
								TEXT ("Warning"), MB_ICONWARNING|MB_SETFOREGROUND|MB_OK|MB_APPLMODAL) ;
							}
						// Get the number of turns for a game in the edit control.
						GetWindowText (GetDlgItem (w_handle, IDC_EDIT_TURNS), tmp_unicode_string, MAX_STRINGLEN) ;
						number_of_turns = (int) _wtoi (tmp_unicode_string) ;
						if (number_of_turns < 5) {
							number_of_turns = 5 ;
							MessageBox (w_handle, TEXT ("Number of turns for a game is lower than 5. Let me set it at 5 otherwise it won't be worth to play !"),
								TEXT ("Warning"), MB_ICONWARNING|MB_SETFOREGROUND|MB_OK|MB_APPLMODAL) ;
							}
						// Save into persistent preferences.
						SavePrefs () ;
						}																				// End of if (MessageBox (w_handle, TEXT ("This will abort t

					// Close the modal dialog.
					EndDialog (w_handle, TRUE) ;
					// Resize and refresh the board window to take into account the
					// new number of turns possible to display in the window.
					ResizeBoardWindow (board_window_handle, (LPARAM) 0) ;
					return (TRUE) ;
					break ;																		// End of IDB_OPTIONSOK

				case IDB_OPTIONSCANCEL:
					// Close the modal dialog
					EndDialog (w_handle, TRUE) ;
					return (TRUE) ;
					break ;																		// End of IDB_OPTIONSCANCEL


				case IDC_BUTTON_CHANGE_DIC: {
					OPENFILENAME of ;
					TCHAR tmp_file_name[256] = TEXT ("\0") ;
					const LPTSTR open_filter = TEXT ("Dictionaries (*.bin)\0*.bin\0All Documents (*.*)\0*.*\0\0") ;

					memset (&of, 0, sizeof (of)) ;
					of.lStructSize = sizeof (of) ;
					of.hwndOwner = NULL ;
					of.lpstrFile = tmp_file_name ;
					of.nMaxFile = 256 ;
					of.lpstrFilter = open_filter ;
					of.Flags = OFN_FILEMUSTEXIST|OFN_PATHMUSTEXIST ;

					if (GetOpenFileName (&of)) {
						if (!CheckValidDictionary (tmp_file_name)) {
							MessageBox (w_handle, TEXT ("Unable to load this file as a valid dictionary. Current dictionary remains unchanged."), TEXT ("Error"),
								MB_SETFOREGROUND|MB_OK|MB_ICONERROR|MB_APPLMODAL) ;
							}														// End of if (!CheckValidDictionary (...
						else
							SetWindowText (GetDlgItem (w_handle, IDC_STATIC_DICNAME), tmp_file_name) ;
						}											// End of if (GetOpenFileName (&of))
					return (TRUE) ;
					}												// End of local block for IDC_BUTTON_CHANGE_DIC
					break ;									// End of IDC_BUTTON_CHANGE_DIC
				}																						// End of switch (wm_id)
			break ;																				// End of WM_COMMAND
		}																								// End switch (message)

	// Tells that we did not handle the message
	// Force default handling by the system (because we're
	// in the case of a dialog handler). Calling DefWindowProc is only the
	// protocol for window handlers
  return (FALSE) ;
}




LRESULT CALLBACK MainWindowHandler (HWND w_handle, UINT message, WPARAM w_param,
																		LPARAM l_param)
{
  int wm_id, wm_event ;

  wm_id    = LOWORD (w_param) ;
  wm_event = HIWORD (w_param) ;

	switch (message) {
		case WM_CREATE: {
			SHMENUBARINFO mbi ;
			WNDCLASS wc ;

			// Create the menu
			memset (&mbi, 0, sizeof (SHMENUBARINFO)) ;
			mbi.cbSize = sizeof (SHMENUBARINFO) ;
			mbi.hwndParent = w_handle ;
			mbi.nToolBarId = IDMB_MAIN_MENU ;
			mbi.hInstRes = app_instance_handle ;
			mbi.nBmpId = 0 ;
			mbi.cBmpImages = 0 ;	
 			SHCreateMenuBar (&mbi) ;
			main_menubar_hd = mbi.hwndMB ;				// Retrieve the handle of the created menubar.

			main_dialog_hd = CreateDialog (app_instance_handle,
				MAKEINTRESOURCE(IDD_DIALOG_MAIN), w_handle, (DLGPROC) MainDlgHandler) ;

			// Create the board window that is a child window of the main window.
			wc.style = 0 ;
			wc.lpfnWndProc = (WNDPROC) BoardWindowHandler ;
			wc.cbClsExtra = 0 ;
			wc.cbWndExtra = 0 ;
			wc.hIcon = NULL ;
			wc.hInstance = app_instance_handle ;
			wc.hCursor = NULL ;
			wc.hbrBackground = NULL ;
			wc.lpszMenuName = NULL ;
			wc.lpszClassName = CLIENT_WINDOW_CLASS_NAME ;
			RegisterClass (&wc) ;

			board_window_handle = CreateWindow (CLIENT_WINDOW_CLASS_NAME, TEXT (""),
				WS_VISIBLE|WS_VSCROLL|WS_CHILD, CW_USEDEFAULT, CW_USEDEFAULT,
				CW_USEDEFAULT, CW_USEDEFAULT,	w_handle, NULL, app_instance_handle, NULL) ;

			ShowWindow (main_dialog_hd, SW_SHOW) ;
			UpdateWindow (main_dialog_hd) ;
			ShowWindow (board_window_handle, SW_SHOW) ;
			UpdateWindow (board_window_handle) ;
			return (0) ;
			}													// End of local block for WM_CREATE
			break ;										// End of WM_CREATE

		case WM_SETTINGCHANGE :
			switch (w_param) {
				case SPI_SETSIPINFO:
					// The state of the SIP changed, so force board resizing.
					ResizeBoardWindow (board_window_handle, l_param) ;
					// Don;t return yet, let the system finish to process the message.
					break ;													// End of SPI_SETSIPINFO
				}																	// End of switch (w_param)
			break ;															// End of WM_SETTINGCHANGE 

		case WM_COMMAND:
			switch (wm_id) {
				case IDMI_NEW_GAME:
					StartNewGame () ;
					return (0) ;
					break ;													// End of IDC_BUTTON_NEW_GAME

				case IDMI_GAME_ABORTGAME:
					if (current_game != NULL) {
						if (MessageBox (w_handle, TEXT ("This will abort the current game. Continue ?"),
								TEXT ("Warning"), MB_SETFOREGROUND|MB_YESNO|MB_ICONQUESTION|MB_APPLMODAL) == IDYES) {
							swprintf (tmp_unicode_string,
								TEXT ("So bad ! The solution was '%S'."), current_game->solution) ;
							MessageBox (w_handle, tmp_unicode_string, TEXT ("Game over"), 
								MB_OK|MB_ICONINFORMATION|MB_APPLMODAL|MB_SETFOREGROUND) ;
							TerminateGame () ;
							}
						}
					return (0) ;
					break ;

				case IDMI_GAME_OPTIONS:
					// Open a modal dialog
					DialogBox (app_instance_handle, MAKEINTRESOURCE (IDD_DIALOG_OPTIONS), w_handle,
						(DLGPROC) OptionDialogHandler) ;
					// Force screen refresh
					InvalidateRect (w_handle, NULL, FALSE) ;
					return (0) ;
					break ;																	// End of IDMI_OPTIONS

				case IDMI_HELP:
					CreateProcess (TEXT("peghelp"), TEXT("PocketMotus.htm"), NULL, NULL,
						FALSE, 0, NULL, NULL, NULL, NULL) ;
					return (0) ;
					break ;												// End of IDMI_HELP

				case IDMI_EXIT:
					// Just triggers a WM_QUIT message that will cause the exit of the mainloop.
					PostMessage (w_handle, WM_CLOSE, 0, 0L) ;
					return (0) ;
					break ;												// End of IDMI_EXIT

				case IDMI_SOFT_HINT:
				case IDMI_HARD_HINT:
					GiveHint (wm_id) ;
					return (0) ;
					break ;												// End of IDMI_SOFT_HINT and IDMI_HARD_HINT
				}																// End for switch (wm_id)
			break ;														// End WM_COMMAND

		case WM_CLOSE:
		// Just triggers a WM_QUIT message that will make us exiting the mainloop.
		PostQuitMessage (0) ;
		return (0) ;
		break ;															// End of WM_CLOSE
		}																		// End for switch (message)

	// Remember, window handler must call DefWindowProc to handle
	// the remaining messages. Returning TRUE or FALSE is only the protocol
	// for dialog handlers !
	return (DefWindowProc (w_handle, message, w_param, l_param)) ;
}





int WINAPI WinMain (HINSTANCE i_handle, HINSTANCE prev_i_handle,
		    LPTSTR cmd_line, int w_aspect)
{
	MSG msg ;
	WNDCLASS wc ;
	HWND w_handle ;
	SIPINFO si ;
	int iDelta ;
	SYSTEMTIME sys_time ;

	app_instance_handle = i_handle ;

	// Check if the program is already running. If yes then focus on the window
	w_handle = FindWindow (MAIN_WINDOW_CLASS_NAME, NULL) ;
	if (w_handle) {
		SetForegroundWindow (w_handle) ;
		return (0) ;
	}

	// Loads the background bitmap
	background_bm_handle =
		(HBITMAP) LoadBitmap (app_instance_handle,
		MAKEINTRESOURCE (IDB_BITMAP_BACKGROUND)) ;
  background_brush = CreatePatternBrush (background_bm_handle) ;

	LoadPrefs () ;

	// Register the window.
	if (! prev_i_handle) {
		wc.style = CS_HREDRAW|CS_VREDRAW ;
		wc.lpfnWndProc = (WNDPROC) MainWindowHandler ;
		wc.cbClsExtra = 0 ;
    wc.cbWndExtra = 0 ;
    wc.hIcon = (HICON) LoadIcon (app_instance_handle, MAKEINTRESOURCE (IDI_ICON_MOTUS)) ;
    wc.hInstance = i_handle ;
    wc.hCursor = NULL ;
    wc.hbrBackground = background_brush ;
    wc.lpszMenuName = NULL ;
    wc.lpszClassName = MAIN_WINDOW_CLASS_NAME ;
    RegisterClass (&wc) ;
  }

	// Loads the DLL associated with common controls.
	InitCommonControls () ;

  // Create the window. Consider the menu at the bottom and the SIP
	si.cbSize = sizeof (si) ;
	SHSipInfo (SPI_GETSIPINFO, 0, &si, 0) ;
	iDelta = (si.fdwFlags & SIPF_ON) ? 0 : MENU_HEIGHT ;
  w_handle = CreateWindow (MAIN_WINDOW_CLASS_NAME, TEXT ("Pocket Motus V1.1"),
			   WS_VISIBLE|WS_CLIPCHILDREN|WS_CLIPSIBLINGS, CW_USEDEFAULT, CW_USEDEFAULT,
			   si.rcVisibleDesktop.right - si.rcVisibleDesktop.left,
				 si.rcVisibleDesktop.bottom - si.rcVisibleDesktop.top - iDelta,
				 NULL, NULL, i_handle, NULL) ;

	SendMessage (w_handle, WM_SETICON, ICON_SMALL,
		(LPARAM) LoadImage (app_instance_handle,
		MAKEINTRESOURCE(IDI_ICON_MOTUS), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR)) ;

	// Finally display the main window
	main_window_handle = w_handle ;
  ShowWindow (w_handle, w_aspect) ;
  UpdateWindow (w_handle) ;

	// Initialize the randomizer
  GetLocalTime (&sys_time) ;
	// wMilliseconds is not reliable on my Ipaq, so use something else !
  srand ((unsigned int) (sys_time.wSecond ^ sys_time.wMinute)) ;

	// Load the eventual dictionary.
	if (automaton_size == 0) {
		if (!LoadDictionary (dic_name))
			MessageBox (main_window_handle, TEXT ("Unable to load dictionary. You may select a valid one later in the game options."), TEXT ("Warning"),
			MB_SETFOREGROUND|MB_OK|MB_ICONWARNING|MB_APPLMODAL) ;
		}

  // Entering main message loop
  while (GetMessage (&msg, NULL, 0, 0)) {
		if ((main_dialog_hd == NULL) || (!IsDialogMessage (main_dialog_hd, &msg))) {
			TranslateMessage (&msg) ;
			DispatchMessage (&msg) ;
			}																// End of if ((main_dialog_hd == NULL) || ...
		}																	// End of while (GetMessage (&msg, NULL, 0, 0))
  
	// In any case, attempt to kill the timers if some exist.
	KillTimer (main_window_handle, STEP_TIMER_ID) ;
	KillTimer (main_window_handle, GLOBAL_TIMER_ID) ;
	DeleteObject (background_brush) ;
	DeleteObject (background_bm_handle) ;
	UnloadDictionary () ;
  return (msg.wParam) ;
}
